"""Utilities/Helpers for defaults and schemas"""
from os import path
from six import string_types
from voluptuous import Any, Boolean, Coerce, Optional

# Elasticsearch versions supported
def version_max():
    """
    :returns: The maximum Elasticsearch version Curator supports: ``(8, 99, 99)``
    """
    return (8, 99, 99)
def version_min():
    """
    :returns: The minimum Elasticsearch version Curator supports: ``(8, 99, 99)``
    """
    return (8, 0, 0)

# Default Config file location
def config_file():
    """
    :returns: The default configuration file location:
        ``path.join(path.expanduser('~'), '.curator', 'curator.yml')``
    """
    return path.join(path.expanduser('~'), '.curator', 'curator.yml')

# Default filter patterns (regular expressions)
def regex_map():
    """
    :returns: A dictionary of pattern filter 'kind's with their associated regular expression:
        ``{'timestring': r'^.*{0}.*$', 'regex': r'{0}', 'prefix': r'^{0}.*$', 'suffix': r'^.*{0}$'}``
    """
    return {
        'timestring': r'^.*{0}.*$',
        'regex': r'{0}',
        'prefix': r'^{0}.*$',
        'suffix': r'^.*{0}$',
    }

def date_regex():
    """
    :returns: A dictionary/map of the strftime string characters and their string lengths:
        ``{'Y':'4', 'G':'4', 'y':'2', 'm':'2', 'W':'2', 'V':'2', 'U':'2', 'd':'2', 'H':'2', 'M':'2', 'S':'2', 'j':'3'}``
    """
    return {
        'Y': '4',
        'G': '4',
        'y': '2',
        'm': '2',
        'W': '2',
        'V': '2',
        'U': '2',
        'd': '2',
        'H': '2',
        'M': '2',
        'S': '2',
        'j': '3',
    }

# Actions

def cluster_actions():
    """
    :returns: A list of supported cluster actions (right now, that's only ``['cluster_routing']``)
    """
    return ['cluster_routing']

def index_actions():
    """
    :returns: The list of supported index actions:
        ``[ 'alias', 'allocation', 'close', 'create_index', 'delete_indices', 'forcemerge', 'index_settings', 'open', 'reindex', 'replicas', 'rollover', 'shrink', 'snapshot']``
    """
    return [
        'alias',
        'allocation',
        'close',
        'cold2frozen',
        'create_index',
        'delete_indices',
        'forcemerge',
        'index_settings',
        'open',
        'reindex',
        'replicas',
        'rollover',
        'shrink',
        'snapshot',
    ]

def snapshot_actions():
    """
    :returns: The list of supported snapshot actions: ``['delete_snapshots', 'restore']``
    """
    return ['delete_snapshots', 'restore']

def all_actions():
    """
    :returns: A sorted list of all supported actions: cluster, index, and snapshot
    """
    return sorted(cluster_actions() + index_actions() + snapshot_actions())

def index_filtertypes():
    """
    :returns: The list of supported index filter types:
        ``['alias', 'allocated', 'age', 'closed', 'count', 'empty', 'forcemerged', 'ilm', 'kibana', 'none', 'opened', 'pattern', 'period', 'space', 'shards', 'size']``
    """

    return [
        'alias',
        'allocated',
        'age',
        'closed',
        'count',
        'empty',
        'forcemerged',
        'ilm',
        'kibana',
        'none',
        'opened',
        'pattern',
        'period',
        'space',
        'shards',
        'size',
    ]

def snapshot_filtertypes():
    """
    :returns: The list of supported snapshot filter types: ``['age', 'count', 'none', 'pattern', 'period', 'state']``
    """
    return ['age', 'count', 'none', 'pattern', 'period', 'state']

def all_filtertypes():
    """
    :returns: A sorted list of all supported filter types (both snapshot and index)
    """
    return sorted(list(set(index_filtertypes() + snapshot_filtertypes())))

def default_options():
    """
    :returns: The default values for these options:
        ``{'allow_ilm_indices': False, 'continue_if_exception': False, 'disable_action': False, 'ignore_empty_list': False, 'timeout_override': None}``
    """
    return {
        'allow_ilm_indices': False,
        'continue_if_exception': False,
        'disable_action': False,
        'ignore_empty_list': False,
        'timeout_override': None,
    }

def default_filters():
    """
    If no filters are set, add a 'none' filter

    :returns: {'filters': [{'filtertype': 'none'}]}
    """
    return {'filters': [{'filtertype': 'none'}]}

def structural_filter_elements():
    """
    :returns: Barebones schemas for initial validation of filters
    """
    # pylint: disable=E1120
    return {
        Optional('aliases'): Any(list, *string_types),
        Optional('allocation_type'): Any(*string_types),
        Optional('count'): Coerce(int),
        Optional('date_from'): Any(None, *string_types),
        Optional('date_from_format'): Any(None, *string_types),
        Optional('date_to'): Any(None, *string_types),
        Optional('date_to_format'): Any(None, *string_types),
        Optional('direction'): Any(*string_types),
        Optional('disk_space'): float,
        Optional('epoch'): Any(Coerce(int), None),
        Optional('exclude'): Any(None, bool, int, *string_types),
        Optional('field'): Any(None, *string_types),
        Optional('intersect'): Any(None, bool, int, *string_types),
        Optional('key'): Any(*string_types),
        Optional('kind'): Any(*string_types),
        Optional('max_num_segments'): Coerce(int),
        Optional('number_of_shards'): Coerce(int),
        Optional('pattern'): Any(*string_types),
        Optional('period_type'): Any(*string_types),
        Optional('reverse'): Any(None, bool, int, *string_types),
        Optional('range_from'): Coerce(int),
        Optional('range_to'): Coerce(int),
        Optional('shard_filter_behavior'): Any(*string_types),
        Optional('size_behavior'): Any(*string_types),
        Optional('size_threshold'): Any(Coerce(float)),
        Optional('source'): Any(*string_types),
        Optional('state'): Any(*string_types),
        Optional('stats_result'): Any(None, *string_types),
        Optional('timestring'): Any(None, *string_types),
        Optional('threshold_behavior'): Any(*string_types),
        Optional('unit'): Any(*string_types),
        Optional('unit_count'): Coerce(int),
        Optional('unit_count_pattern'): Any(*string_types),
        Optional('use_age'): Boolean(),
        Optional('value'): Any(int, float, bool, *string_types),
        Optional('week_starts_on'): Any(None, *string_types),
    }



"""Action Option Schema definitions"""
from six import string_types
from voluptuous import All, Any, Boolean, Coerce, Optional, Range, Required

# pylint: disable=line-too-long,missing-docstring
# pylint: disable=E1120


def allocation_type():
    """
    :returns: ``{Optional('allocation_type', default='require'): All(Any(*string_types), Any('require', 'include', 'exclude'))}``
    """
    return {Optional('allocation_type', default='require'): All(Any(*string_types), Any('require', 'include', 'exclude'))}

def allow_ilm_indices():
    """
    :returns: ``{Optional('allow_ilm_indices', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('allow_ilm_indices', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def conditions():
    """
    :returns: ``{Optional('conditions'): {Optional('max_age'): Any(*string_types), Optional('max_docs'): Coerce(int), Optional('max_size'): Any(*string_types)}}``
    """
    return {Optional('conditions'): {Optional('max_age'): Any(*string_types), Optional('max_docs'): Coerce(int), Optional('max_size'): Any(*string_types)}}

def continue_if_exception():
    """
    :returns: ``{Optional('continue_if_exception', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('continue_if_exception', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def count():
    """
    :returns: ``{Required('count'): All(Coerce(int), Range(min=0, max=10))}``
    """
    return {Required('count'): All(Coerce(int), Range(min=0, max=10))}

def delay():
    """
    :returns: ``{Optional('delay', default=0): All(Coerce(float), Range(min=0.0, max=3600.0))}``
    """
    return {Optional('delay', default=0): All(Coerce(float), Range(min=0.0, max=3600.0))}

def c2f_index_settings():
    """
    Only for the :py:class:`~.curator.actions.Cold2Frozen` action

    :returns: ``{Optional('index_settings'):  Any(None, dict)}``
    """
    return {Optional('index_settings', default=None):  Any(None, dict)}

def c2f_ignore_index_settings():
    """
    Only for the :py:class:`~.curator.actions.Cold2Frozen` action

    :returns: ``{Optional('ignore_index_settings'):  Any(None, list)}``
    """
    return {Optional('ignore_index_settings', default=None):  Any(None, list)}

def copy_aliases():
    """
    :returns: ``{Optional('copy_aliases', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('copy_aliases', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def delete_after():
    """
    :returns: ``{Optional('delete_after', default=True): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('delete_after', default=True): Any(bool, All(Any(*string_types), Boolean()))}

def delete_aliases():
    """
    :returns: ``{Optional('delete_aliases', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('delete_aliases', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def skip_flush():
    """
    :returns: ``{Optional('skip_flush', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('skip_flush', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def disable_action():
    """
    :returns: ``{Optional('disable_action', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('disable_action', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def extra_settings():
    """
    :returns: ``{Optional('extra_settings', default={}): dict}``
    """
    return {Optional('extra_settings', default={}): dict}

def ignore_empty_list():
    """
    :returns: ``{Optional('ignore_empty_list', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('ignore_empty_list', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def ignore_existing():
    """
    :returns: ``{Optional('ignore_existing', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('ignore_existing', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def ignore_unavailable():
    """
    :returns: ``{Optional('ignore_unavailable', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('ignore_unavailable', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def include_aliases():
    """
    :returns: ``{Optional('include_aliases', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('include_aliases', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def include_global_state(action):
    """
    :returns: ``{Optional('include_global_state', default=default): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    default = False
    if action == 'snapshot':
        default = True
    return {Optional('include_global_state', default=default): Any(bool, All(Any(*string_types), Boolean()))}

def index_settings():
    """
    :returns: ``{Required('index_settings'): {'index': dict}}``
    """
    return {Required('index_settings'): {'index': dict}}

def indices():
    """
    :returns: ``{Optional('indices', default=None): Any(None, list)}``
    """
    return {Optional('indices', default=None): Any(None, list)}

def key():
    """
    :returns: ``{Required('key'): Any(*string_types)}``
    """
    return {Required('key'): Any(*string_types)}

def max_num_segments():
    """
    :returns: ``{Required('max_num_segments'): All(Coerce(int), Range(min=1, max=32768))}``
    """
    return {Required('max_num_segments'): All(Coerce(int), Range(min=1, max=32768))}

# pylint: disable=unused-argument
def max_wait(action):
    """
    :returns: ``{Optional('max_wait', default=defval): Any(-1, Coerce(int), None)}``
    """
    # The separation is here in case I want to change defaults later...
    defval = -1
    # if action in ['allocation', 'cluster_routing', 'replicas']:
    #     defval = -1
    # elif action in ['restore', 'snapshot', 'reindex', 'shrink']:
    #     defval = -1
    return {Optional('max_wait', default=defval): Any(-1, Coerce(int), None)}

def migration_prefix():
    """
    :returns: ``{Optional('migration_prefix', default=''): Any(None, *string_types)}``
    """
    return {Optional('migration_prefix', default=''): Any(None, *string_types)}

def migration_suffix():
    """
    :returns: ``{Optional('migration_suffix', default=''): Any(None, *string_types)}``
    """
    return {Optional('migration_suffix', default=''): Any(None, *string_types)}

def name(action):
    """
    :returns: The proper name based on what action it is:
        ``alias``, ``create_index``, ``rollover``: ``{Required('name'): Any(*string_types)}``
        ``snapshot``: ``{Optional('name', default='curator-%Y%m%d%H%M%S'): Any(*string_types)}``
        ``restore``: ``{Optional('name'): Any(*string_types)}``
    """
    if action in ['alias', 'create_index', 'rollover']:
        return {Required('name'): Any(*string_types)}
    if action == 'snapshot':
        return {Optional('name', default='curator-%Y%m%d%H%M%S'): Any(*string_types)}
    if action == 'restore':
        return {Optional('name'): Any(*string_types)}


def new_index():
    return {Optional('new_index', default=None): Any(None, *string_types)}

def node_filters():
    """
    :returns: A :py:class:`voluptuous.schema_builder.Schema` object. See code for more details.
    """
    return {
        Optional('node_filters', default={}): {
            Optional('permit_masters', default=False): Any(bool, All(Any(*string_types), Boolean())),
            Optional('exclude_nodes', default=[]): Any(list, None)
        }
    }

def number_of_replicas():
    """
    :returns: ``{Optional('number_of_replicas', default=1): All(Coerce(int), Range(min=0, max=10))}``
    """
    return {Optional('number_of_replicas', default=1): All(Coerce(int), Range(min=0, max=10))}

def number_of_shards():
    """
    :returns: ``{Optional('number_of_shards', default=1): All(Coerce(int), Range(min=1, max=99))}``
    """
    return {Optional('number_of_shards', default=1): All(Coerce(int), Range(min=1, max=99))}

def partial():
    """
    :returns: ``{Optional('partial', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('partial', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def post_allocation():
    """
    :returns: A :py:class:`voluptuous.schema_builder.Schema` object. See code for more details.
    """
    return {
        Optional('post_allocation', default={}):
            Any(
                {},
                All(
                    {
                        Required('allocation_type', default='require'): All(Any(*string_types), Any('require', 'include', 'exclude')),
                        Required('key'): Any(*string_types),
                        Required('value', default=None): Any(None, *string_types)
                    }
                )
            )
    }

def preserve_existing():
    """
    :returns: ``{Optional('preserve_existing', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('preserve_existing', default=False): Any(bool, All(Any(*string_types), Boolean()))}

def refresh():
    """
    :returns: ``{Optional('refresh', default=True): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('refresh', default=True): Any(bool, All(Any(*string_types), Boolean()))}

def remote_certificate():
    """
    :returns: ``{Optional('remote_certificate', default=None): Any(None, *string_types)}``
    """
    return {Optional('remote_certificate', default=None): Any(None, *string_types)}

def remote_client_cert():
    """
    :returns: ``{Optional('remote_client_cert', default=None): Any(None, *string_types)}``
    """
    return {Optional('remote_client_cert', default=None): Any(None, *string_types)}

def remote_client_key():
    """
    :returns: ``{Optional('remote_client_key', default=None): Any(None, *string_types)}``
    """
    return {Optional('remote_client_key', default=None): Any(None, *string_types)}

def remote_filters():
    """
    :returns: A :py:class:`voluptuous.schema_builder.Schema` object. See code for more details.
    """
    # This is really just a basic check here.  The real check is in the
    # validate_actions() method in utils.py
    return {
        Optional(
            'remote_filters',
            default=[
                {
                    'filtertype': 'pattern',
                    'kind': 'regex',
                    'value': '.*',
                    'exclude': True,
                }
            ]
        ): Any(list, None)
    }

def rename_pattern():
    """
    :returns: ``{Optional('rename_pattern'): Any(*string_types)}``
    """
    return {Optional('rename_pattern'): Any(*string_types)}

def rename_replacement():
    """
    :returns: ``{Optional('rename_replacement'): Any(*string_types)}``
    """
    return {Optional('rename_replacement'): Any(*string_types)}

def repository():
    """
    :returns: ``{Required('repository'): Any(*string_types)}``
    """
    return {Required('repository'): Any(*string_types)}

def request_body():
    """
    :returns: A :py:class:`voluptuous.schema_builder.Schema` object. See code for more details.
    """
    return {
        Required('request_body'): {
            Optional('conflicts'): Any('proceed', 'abort'),
            Optional('max_docs'): Coerce(int),
            Required('source'): {
                Required('index'): Any(Any(*string_types), list),
                Optional('query'): dict,
                Optional('remote'): {
                    Optional('host'): Any(*string_types),
                    Optional('username'): Any(*string_types),
                    Optional('password'): Any(*string_types),
                    Optional('socket_timeout'): Any(*string_types),
                    Optional('connect_timeout'): Any(*string_types),
                    Optional('headers'): Any(*string_types),
                },
                Optional('size'): Coerce(int),
                Optional('_source'): Any(bool, Boolean()),
            },
            Required('dest'): {
                Required('index'): Any(*string_types),
                Optional('version_type'): Any('internal', 'external', 'external_gt', 'external_gte'),
                Optional('op_type'): Any(*string_types),
                Optional('pipeline'): Any(*string_types),
            },
            Optional('script'): {
                Optional('source'): Any(*string_types),
                Optional('lang'): Any('painless', 'expression', 'mustache', 'java')
            },
        }
    }

def requests_per_second():
    """
    :returns: ``{Optional('requests_per_second', default=-1): Any(-1, Coerce(int), None)}``
    """
    return {Optional('requests_per_second', default=-1): Any(-1, Coerce(int), None)}

def retry_count():
    """
    :returns: ``{Optional('retry_count', default=3): All(Coerce(int), Range(min=0, max=100))}``
    """
    return {Optional('retry_count', default=3): All(Coerce(int), Range(min=0, max=100))}

def retry_interval():
    """
    :returns: ``{Optional('retry_interval', default=120): All(Coerce(int), Range(min=1, max=600))}``
    """
    return {Optional('retry_interval', default=120): All(Coerce(int), Range(min=1, max=600))}

def routing_type():
    """
    :returns: ``{Required('routing_type'): Any('allocation', 'rebalance')}``
    """
    return {Required('routing_type'): Any('allocation', 'rebalance')}

def cluster_routing_setting():
    """
    :returns: ``{Required('setting'): Any('enable')}``
    """
    return {Required('setting'): Any('enable')}

def cluster_routing_value():
    """
    :returns: ``{Required('value'): Any('all', 'primaries', 'none', 'new_primaries', 'replicas')}``
    """
    return {Required('value'): Any('all', 'primaries', 'none', 'new_primaries', 'replicas')}

def shrink_node():
    """
    :returns: ``{Required('shrink_node'): Any(*string_types)}``
    """
    return {Required('shrink_node'): Any(*string_types)}

def shrink_prefix():
    """
    :returns: ``{Optional('shrink_prefix', default=''): Any(None, *string_types)}``
    """
    return {Optional('shrink_prefix', default=''): Any(None, *string_types)}

def shrink_suffix():
    """
    :returns: ``{Optional('shrink_suffix', default='-shrink'): Any(None, *string_types)}``
    """
    return {Optional('shrink_suffix', default='-shrink'): Any(None, *string_types)}

def skip_repo_fs_check():
    """
    :returns: ``{Optional('skip_repo_fs_check', default=True): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('skip_repo_fs_check', default=True): Any(bool, All(Any(*string_types), Boolean()))}

def slices():
    """
    :returns: ``{Optional('slices', default=1): Any(All(Coerce(int), Range(min=1, max=500)), None)}``
    """
    return {Optional('slices', default=1): Any(All(Coerce(int), Range(min=1, max=500)), None)}

def timeout(action):
    """
    :returns: ``{Optional('timeout', default=defval): Any(Coerce(int), None)}``
    """
    # if action == 'reindex':
    defval = 60
    return {Optional('timeout', default=defval): Any(Coerce(int), None)}

def timeout_override(action):
    """
    :returns: ``{Optional('timeout_override', default=defval): Any(Coerce(int), None)}`` where
        ``defval`` is determined by the action.

        ``['forcemerge', 'restore', 'snapshot']`` = ``21600``

        ``close`` = ``180``

        ``delete_snapshots`` = ``300``
    """
    if action in ['forcemerge', 'restore', 'snapshot']:
        defval = 21600
    elif action == 'close':
        defval = 180
    elif action == 'delete_snapshots':
        defval = 300
    else:
        defval = None
    return {Optional('timeout_override', default=defval): Any(Coerce(int), None)}

def value():
    """
    :returns: ``{Required('value', default=None): Any(None, *string_types)}``
    """
    return {Required('value', default=None): Any(None, *string_types)}

def wait_for_active_shards(action):
    """
    :returns: ``{Optional('wait_for_active_shards', default=defval): Any(Coerce(int), 'all', None)}``
        where ``defval`` defaults to 0, but changes to 1 for the ``reindex`` and ``shrink`` actions.
    """
    defval = 0
    if action in ['reindex', 'shrink']:
        defval = 1
    return {Optional('wait_for_active_shards', default=defval): Any(Coerce(int), 'all', None)}

def wait_for_completion(action):
    """
    :returns: ``{Optional('wait_for_completion', default=defval): Any(bool, All(Any(*string_types), Boolean()))}``
        where ``defval`` defaults to True, but changes to False if action is ``allocation``,
        ``cluster_routing``, or ``replicas``.
    """
    # if action in ['cold2frozen', 'reindex', 'restore', 'snapshot']:
    defval = True
    if action in ['allocation', 'cluster_routing', 'replicas']:
        defval = False
    return {Optional('wait_for_completion', default=defval): Any(bool, All(Any(*string_types), Boolean()))}

def wait_for_rebalance():
    """
    :returns: ``{Optional('wait_for_rebalance', default=True): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('wait_for_rebalance', default=True): Any(bool, All(Any(*string_types), Boolean()))}

def wait_interval(action):
    """
    :returns: ``{Optional('wait_interval', default=defval): Any(All(Coerce(int), Range(min=minval, max=maxval)), None)}``
        where ``minval`` = ``1``, ``maxval`` = ``30``, and ``defval`` is ``3``, unless the action
        is one of ``['restore', 'snapshot', 'reindex', 'shrink']``, and then ``defval`` is ``9``.
    """
    minval = 1
    maxval = 30
    # if action in ['allocation', 'cluster_routing', 'replicas']:
    defval = 3
    if action in ['restore', 'snapshot', 'reindex', 'shrink']:
        defval = 9
    return {Optional('wait_interval', default=defval): Any(All(Coerce(int), Range(min=minval, max=maxval)), None)}

def warn_if_no_indices():
    """
    :returns: ``{Optional('warn_if_no_indices', default=False): Any(bool, All(Any(*string_types), Boolean()))}``
    """
    return {Optional('warn_if_no_indices', default=False): Any(bool, All(Any(*string_types), Boolean()))}



"""Shrink Index Singleton"""
import click
from curator.cli_singletons.object_class import CLIAction
from curator.cli_singletons.utils import get_width, json_to_dict, validate_filter_json

# pylint: disable=line-too-long
@click.command(context_settings=get_width())
@click.option('--shrink_node', default='DETERMINISTIC', type=str, help='Named node, or DETERMINISTIC', show_default=True)
@click.option('--node_filters', help='JSON version of node_filters (see documentation)', callback=json_to_dict)
@click.option('--number_of_shards', default=1, type=int, help='Shrink to this many shards per index')
@click.option('--number_of_replicas', default=1, type=int, help='Number of replicas for the target index', show_default=True)
@click.option('--shrink_prefix', type=str, help='Prefix for the target index name')
@click.option('--shrink_suffix', default='-shrink', type=str, help='Suffix for the target index name', show_default=True)
@click.option('--copy_aliases', is_flag=True, help='Copy each source index aliases to target index')
@click.option('--delete_after/--no-delete_after', default=True, help='Delete source index after shrink', show_default=True)
@click.option('--post_allocation', help='JSON version of post_allocation (see documentation)', callback=json_to_dict)
@click.option('--extra_settings', help='JSON version of extra_settings (see documentation)', callback=json_to_dict)
@click.option('--wait_for_active_shards', default=1, type=int, help='Wait for number of active shards before continuing')
@click.option('--wait_for_rebalance/--no-wait_for_rebalance', default=True, help='Wait for rebalance to complete')
@click.option('--wait_for_completion/--no-wait_for_completion', default=True, help='Wait for the shrink to complete')
@click.option('--wait_interval', default=9, type=int, help='Seconds to wait between completion checks.')
@click.option('--max_wait', default=-1, type=int, help='Maximum number of seconds to wait_for_completion')
@click.option('--ignore_empty_list', is_flag=True, help='Do not raise exception if there are no actionable indices')
@click.option('--allow_ilm_indices/--no-allow_ilm_indices', help='Allow Curator to operate on Index Lifecycle Management monitored indices.', default=False, show_default=True)
@click.option('--filter_list', callback=validate_filter_json, help='JSON array of filters selecting indices to act on.', required=True)
@click.pass_context
def shrink(
        ctx, shrink_node, node_filters, number_of_shards, number_of_replicas, shrink_prefix,
        shrink_suffix, copy_aliases, delete_after, post_allocation, extra_settings,
        wait_for_active_shards, wait_for_rebalance, wait_for_completion, wait_interval, max_wait,
        ignore_empty_list, allow_ilm_indices, filter_list
    ):
    """
    Shrink Indices to --number_of_shards
    """
    manual_options = {
        'shrink_node': shrink_node,
        'node_filters': node_filters,
        'number_of_shards': number_of_shards,
        'number_of_replicas': number_of_replicas,
        'shrink_prefix': shrink_prefix,
        'shrink_suffix': shrink_suffix,
        'copy_aliases': copy_aliases,
        'delete_after': delete_after,
        'post_allocation': post_allocation,
        'extra_settings': extra_settings,
        'wait_for_active_shards': wait_for_active_shards,
        'wait_for_rebalance': wait_for_rebalance,
        'wait_for_completion': wait_for_completion,
        'wait_interval': wait_interval,
        'max_wait': max_wait,
        'allow_ilm_indices': allow_ilm_indices,
    }
    # ctx.info_name is the name of the function or name specified in @click.command decorator
    action = CLIAction(ctx.info_name, ctx.obj['config'], manual_options, filter_list, ignore_empty_list)
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])



"""Configuration utilty functions"""
import logging
from copy import deepcopy
import click
from es_client.helpers.utils import prune_nones, ensure_list
from curator.defaults.logging_defaults import config_logging
from curator.validators.schemacheck import SchemaCheck
from curator.logtools import LogInfo, Blacklist

def check_logging_config(config):
    """
    Ensure that the top-level key ``logging`` is in ``config`` before passing it to
    :py:class:`~.curator.validators.schemacheck.SchemaCheck` for value validation.

    :param config: Logging configuration data

    :type config: dict

    :returns: :py:class:`~.curator.validators.schemacheck.SchemaCheck` validated logging
        configuration.
    """

    if not isinstance(config, dict):
        click.echo(
            f'Must supply logging information as a dictionary. '
            f'You supplied: "{config}" which is "{type(config)}"'
            f'Using default logging values.'
        )
        log_settings = {}
    elif not 'logging' in config:
        click.echo('No "logging" setting in supplied configuration.  Using defaults.')
        log_settings = {}
    else:
        if config['logging']:
            log_settings = prune_nones(config['logging'])
        else:
            log_settings = {}
    return SchemaCheck(
        log_settings, config_logging(), 'Logging Configuration', 'logging').result()

def set_logging(log_opts):
    """Configure global logging options

    :param log_opts: Logging configuration data

    :type log_opts: dict

    :rtype: None
    """
    # Set up logging
    loginfo = LogInfo(log_opts)
    logging.root.addHandler(loginfo.handler)
    logging.root.setLevel(loginfo.numeric_log_level)
    _ = logging.getLogger('curator.cli')
    # Set up NullHandler() to handle nested elasticsearch8.trace Logger
    # instance in elasticsearch python client
    logging.getLogger('elasticsearch8.trace').addHandler(logging.NullHandler())
    if log_opts['blacklist']:
        for bl_entry in ensure_list(log_opts['blacklist']):
            for handler in logging.root.handlers:
                handler.addFilter(Blacklist(bl_entry))

def password_filter(data):
    """
    Recursively look through all nested structures of ``data`` for the key ``'password'`` and redact
    the value.

    :param data: Configuration data

    :type data: dict

    :returns: A :py:class:`~.copy.deepcopy` of ``data`` with the value obscured by ``REDACTED``
        if the key is ``'password'``.
    """
    def iterdict(mydict):
        for key, value in mydict.items():
            if isinstance(value, dict):
                iterdict(value)
            elif key == "password":
                mydict.update({"password": "REDACTED"})
        return mydict
    return iterdict(deepcopy(data))



"""Open index action class"""
import logging
from curator.exceptions import ConfigurationError
from curator.helpers.date_ops import parse_date_pattern
from curator.helpers.testers import rollable_alias, verify_client_object
from curator.helpers.utils import report_failure

class Rollover:
    """Rollover Action Class"""
    def __init__(
            self, client, name=None, conditions=None, new_index=None, extra_settings=None,
            wait_for_active_shards=1
        ):
        """
        :param client: A client connection object
        :param name: The name of the single-index-mapped alias to test for rollover conditions.
        :param new_index: A new index name
        :param conditions: Conditions to test
        :param extra_settings: Must be either ``None``, or a dictionary of settings to apply to the new index on rollover. This is used in place of ``settings`` in the Rollover API, mostly because it's already existent in other places here in Curator
        :param wait_for_active_shards: The number of shards expected to be active before returning.

        :type client: :py:class:`~.elasticsearch.Elasticsearch`
        :type name: str
        :type new_index: str
        :type conditions: dict
        :type extra_settings: dict or None
        :type wait_for_active_shards: int
        """
        self.loggit = logging.getLogger('curator.actions.rollover')
        if not isinstance(conditions, dict):
            raise ConfigurationError('"conditions" must be a dictionary')
        else:
            self.loggit.debug('"conditions" is %s', conditions)
        if not isinstance(extra_settings, dict) and extra_settings is not None:
            raise ConfigurationError(
                '"extra_settings" must be a dictionary or None')
        verify_client_object(client)
        #: Object attribute that gets the value of param ``client``.
        self.client = client
        #: Object attribute that gets the value of param ``conditions``.
        self.conditions = conditions
        #: Object attribute that gets the value of param ``extra_settings``.
        self.settings = extra_settings
        #: The :py:func:`~.curator.helpers.date_ops.parse_date_pattern` rendered version of what
        #: was passed as ``new_index``, or else ``None``
        self.new_index = parse_date_pattern(new_index) if new_index else new_index
        #: Object attribute that gets the value of param ``wait_for_active_shards``.
        self.wait_for_active_shards = wait_for_active_shards

        #: Object attribute that gets the value of param ``name``.
        self.name = None
        # Verify that `conditions` and `settings` are good?
        # Verify that `name` is an alias, and is only mapped to one index.
        if rollable_alias(client, name):
            self.name = name
        else:
            raise ValueError(
                f'Unable to perform index rollover with alias '
                f'"{name}". See previous logs for more details.'
            )

    def log_result(self, result):
        """Log the results based on whether the index rolled over or not"""
        dryrun_string = ''
        if result['dry_run']:
            dryrun_string = 'DRY-RUN: '
        self.loggit.debug('%sResult: %s', dryrun_string, result)
        rollover_string = (
            f"{dryrun_string}Old index {result['old_index']} "
            f"rolled over to new index {result['new_index']}"
        )
        # Success is determined by at one condition being True
        success = False
        for k in list(result['conditions'].keys()):
            if result['conditions'][k]:
                success = True
        if result['dry_run'] and success: # log "successful" dry-run
            self.loggit.info(rollover_string)
        elif result['rolled_over']:
            self.loggit.info(rollover_string)
        else:
            msg = (
                f"{dryrun_string}Rollover conditions not met. "
                f"Index {result['old_index']} not rolled over."
            )
            self.loggit.info(msg)

    def doit(self, dry_run=False):
        """
        This exists solely to prevent having to have duplicate code in both  :py:meth:`do_dry_run`
        and :py:meth:`do_action` because :py:meth:`~.elasticsearch.client.IndicesClient.rollover`
        has its own ``dry_run`` flag.
        """
        return self.client.indices.rollover(
            alias=self.name,
            new_index=self.new_index,
            conditions=self.conditions,
            settings=self.settings,
            dry_run=dry_run,
            wait_for_active_shards=self.wait_for_active_shards,
        )

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        self.loggit.info('DRY-RUN MODE.  No changes will be made.')
        self.log_result(self.doit(dry_run=True))

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.IndicesClient.rollover` the index referenced by alias
        :py:attr:`name`
        """
        self.loggit.info('Performing index rollover')
        try:
            self.log_result(self.doit())
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)



"""Helper utilities

The kind that don't fit in testers, getters, date_ops, or converters
"""
import logging
from es_client.helpers.utils import ensure_list
from curator.exceptions import FailedExecution

def chunk_index_list(indices):
    """
    This utility chunks very large index lists into 3KB chunks.
    It measures the size as a csv string, then converts back into a list for the return value.

    :param indices: The list of indices

    :type indices: list

    :returns: A list of lists (each a piece of the original ``indices``)
    :rtype: list
    """
    chunks = []
    chunk = ""
    for index in indices:
        if len(chunk) < 3072:
            if not chunk:
                chunk = index
            else:
                chunk += "," + index
        else:
            chunks.append(chunk.split(','))
            chunk = index
    chunks.append(chunk.split(','))
    return chunks

def report_failure(exception):
    """
    Raise a :py:exc:`~.curator.exceptions.FailedExecution` exception and include the original error
    message.

    :param exception: The upstream exception.

    :type exception: :py:exc:Exception

    :rtype: None
    """
    raise FailedExecution(
        f'Exception encountered.  Rerun with loglevel DEBUG and/or check Elasticsearch logs for'
        f'more information. Exception: {exception}'
    )

def show_dry_run(ilo, action, **kwargs):
    """
    Log dry run output with the action which would have been executed.

    :param ilo: An IndexList Object
    :param action: The ``action`` to be performed.
    :param kwargs: Any other args to show in the log output


    :type ilo: :py:class:`~.curator.indexlist.IndexList`
    :type action: str
    :type kwargs: dict

    :rtype: None
    """
    logger = logging.getLogger(__name__)
    logger.info('DRY-RUN MODE.  No changes will be made.')
    msg = f'(CLOSED) indices may be shown that may not be acted on by action "{action}".'
    logger.info(msg)
    indices = sorted(ilo.indices)
    for idx in indices:
        # Dry runs need index state, so we collect it here if it's not present.
        try:
            index_closed = ilo.index_info[idx]['state'] == 'close'
        except KeyError:
            ilo.get_index_state()
            index_closed = ilo.index_info[idx]['state'] == 'close'
        var = ' (CLOSED)' if index_closed else ''
        msg = f'DRY-RUN: {action}: {idx}{var} with arguments: {kwargs}'
        logger.info(msg)

def to_csv(indices):
    """
    :param indices: A list of indices to act on, or a single value, which could be
        in the format of a csv string already.

    :type indices: list

    :returns: A csv string from a list of indices, or a single value if only one value is present
    :rtype: str
    """
    indices = ensure_list(indices) # in case of a single value passed
    if indices:
        return ','.join(sorted(indices))
    return None



"""Snapshot and Restore action classes"""
import logging
from curator.helpers.getters import get_alias_actions, get_frozen_prefix, get_tier_preference
from curator.helpers.testers import has_lifecycle_name, is_idx_partial, verify_index_list
from curator.helpers.utils import report_failure
from curator.exceptions import CuratorException, FailedExecution, SearchableSnapshotException

class Cold2Frozen:
    """Cold to Frozen Tier Searchable Snapshot Action Class

    For manually migrating snapshots not associated with ILM from the cold tier to the frozen tier.
    """

    DEFAULTS = {
        'index_settings': None,
        'ignore_index_settings': ['index.refresh_interval'],
        'wait_for_completion': True,
    }
    def __init__(self, ilo, **kwargs):
        """
        :param ilo: An IndexList Object
        :param index_settings: (Optional) Settings that should be added to the index when it is
            mounted. If not set, set the ``_tier_preference`` to the tiers available, coldest
            first.
        :param ignore_index_settings: (Optional, array of strings) Names of settings that should
            be removed from the index when it is mounted.
        :param wait_for_completion: Wait for completion before returning.

        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        :type index_settings: dict
        :type ignore_index_settings: list
        :type wait_for_completion: bool
        """
        self.loggit = logging.getLogger('curator.actions.cold2frozen')
        verify_index_list(ilo)
        # Check here and don't bother with the rest of this if there are no
        # indices in the index list.
        ilo.empty_list_check()

        #: The :py:class:`~.curator.indexlist.IndexList` object passed from param ``ilo``
        self.index_list = ilo
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`index_list`
        self.client = ilo.client
        #: Object attribute that contains the :py:func:`~.curator.helpers.utils.to_csv` output of
        #: the indices in :py:attr:`index_list`.
        self.indices = ilo
        #: Object attribute that gets the value of ``index_settings``.
        self.index_settings = None
        #: Object attribute that gets the value of ``ignore_index_settings``.
        self.ignore_index_settings = None
        #: Object attribute that gets the value of param ``wait_for_completion``.
        self.wait_for_completion = None

        # Parse the kwargs into attributes
        self.assign_kwargs(**kwargs)

    def assign_kwargs(self, **kwargs):
        """
        Assign the kwargs to the attribute of the same name with the passed value or the default
        from DEFAULTS
        """
        # Handy little loop here only adds kwargs that exist in DEFAULTS, or the default value.
        # It ignores any non-relevant kwargs
        for key, value in self.DEFAULTS.items():
            if key in kwargs:
                setattr(self, key, kwargs[key])
            else:
                setattr(self, key, value)

    def action_generator(self):
        """Yield a dict for use in :py:meth:`do_action` and :py:meth:`do_dry_run`

        :returns: A generator object containing the settings necessary to migrate indices from cold
            to frozen
        :rtype: dict
        """
        for idx in self.index_list.indices:
            idx_settings = self.client.indices.get(index=idx)[idx]['settings']['index']
            if has_lifecycle_name(idx_settings):
                self.loggit.critical(
                    'Index %s is associated with an ILM policy and this action will never work on '
                    'an index associated with an ILM policy', idx)
                raise CuratorException(f'Index {idx} is associated with an ILM policy')
            if is_idx_partial(idx_settings):
                self.loggit.critical('Index %s is already in the frozen tier', idx)
                raise SearchableSnapshotException('Index is already in frozen tier')
            snap = idx_settings['store']['snapshot']['snapshot_name']
            snap_idx = idx_settings['store']['snapshot']['index_name']
            repo = idx_settings['store']['snapshot']['repository_name']
            aliases = self.client.indices.get(index=idx)[idx]['aliases']

            prefix = get_frozen_prefix(snap_idx, idx)
            renamed = f'{prefix}{snap_idx}'

            if not self.index_settings:
                self.index_settings = {
                    "routing": {
                        "allocation": {
                            "include": {
                                "_tier_preference": get_tier_preference(self.client)
                            }
                        }
                    }
                }
            yield {
                'repository': repo, 'snapshot': snap, 'index': snap_idx,
                'renamed_index': renamed, 'index_settings': self.index_settings,
                'ignore_index_settings': self.ignore_index_settings,
                'storage': 'shared_cache', 'wait_for_completion': self.wait_for_completion,
                'aliases': aliases, 'current_idx': idx
            }

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        self.loggit.info('DRY-RUN MODE.  No changes will be made.')
        for kwargs in self.action_generator():
            aliases = kwargs.pop('aliases')
            current_idx = kwargs.pop('current_idx')
            msg = (
                f'DRY-RUN: cold2frozen: from snapshot {kwargs["snapshot"]} in repository '
                f'{kwargs["repository"]}, mount index {kwargs["index"]} renamed as '
                f'{kwargs["renamed_index"]} with index settings: {kwargs["index_settings"]} '
                f'and ignoring settings: {kwargs["ignore_index_settings"]}. wait_for_completion: '
                f'{kwargs["wait_for_completion"]}. Restore aliases: {aliases}. Current index '
                f'name: {current_idx}'
            )
            self.loggit.info(msg)


    def do_action(self):
        """
        Call :py:meth:`~.elasticsearch.client.SearchableSnapshotsClient.mount` to mount the indices
        in :py:attr:`ilo` in the Frozen tier.

        Verify index looks good

        Call :py:meth:`~.elasticsearch.client.IndicesClient.update_aliases` to update each new
        frozen index with the aliases from the old cold-tier index.

        Verify aliases look good.

        Call :py:meth:`~.elasticsearch.client.IndicesClient.delete` to delete the cold tier index.
        """
        try:
            for kwargs in self.action_generator():
                aliases = kwargs.pop('aliases')
                current_idx = kwargs.pop('current_idx')
                newidx = kwargs['renamed_index']
                # Actually do the mount
                self.loggit.debug('Mounting new index %s in frozen tier...', newidx)
                self.client.searchable_snapshots.mount(**kwargs)
                # Verify it's mounted as a partial now:
                self.loggit.debug('Verifying new index %s is mounted properly...', newidx)
                idx_settings = self.client.indices.get(index=newidx)[newidx]
                if is_idx_partial(idx_settings['settings']['index']):
                    self.loggit.info('Index %s is mounted for frozen tier', newidx)
                else:
                    raise SearchableSnapshotException(
                        f'Index {newidx} not a mounted searchable snapshot')
                # Update Aliases
                alias_names = aliases.keys()
                if not alias_names:
                    self.loggit.warning('No aliases associated with index %s', current_idx)
                else:
                    self.loggit.debug('Transferring aliases to new index %s', newidx)
                    self.client.indices.update_aliases(
                        actions=get_alias_actions(current_idx, newidx, aliases))
                    verify = self.client.indices.get(index=newidx)[newidx]['aliases'].keys()
                    if alias_names != verify:
                        self.loggit.error(
                            'Alias names do not match! %s does not match: %s', alias_names, verify)
                        raise FailedExecution('Aliases failed to transfer to new index')
                # Clean up old index
                self.loggit.debug('Deleting old index: %s', current_idx)
                self.client.indices.delete(index=current_idx)
                self.loggit.info(
                    'Successfully migrated %s to the frozen tier as %s', current_idx, newidx)

        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)



"""Curator Helper Modules"""
from curator.helpers.date_ops import *
from curator.helpers.getters import *
from curator.helpers.testers import *
from curator.helpers.utils import *
from curator.helpers.waiters import *



"""Use __init__ to make these not need to be nested under lowercase.Capital"""
from curator.actions.alias import Alias
from curator.actions.allocation import Allocation
from curator.actions.close import Close
from curator.actions.cluster_routing import ClusterRouting
from curator.actions.cold2frozen import Cold2Frozen
from curator.actions.create_index import CreateIndex
from curator.actions.delete_indices import DeleteIndices
from curator.actions.forcemerge import ForceMerge
from curator.actions.index_settings import IndexSettings
from curator.actions.open import Open
from curator.actions.reindex import Reindex
from curator.actions.replicas import Replicas
from curator.actions.rollover import Rollover
from curator.actions.shrink import Shrink
from curator.actions.snapshot import Snapshot, DeleteSnapshots, Restore

CLASS_MAP = {
    'alias' : Alias,
    'allocation' : Allocation,
    'close' : Close,
    'cluster_routing' : ClusterRouting,
    'cold2frozen': Cold2Frozen,
    'create_index' : CreateIndex,
    'delete_indices' : DeleteIndices,
    'delete_snapshots' : DeleteSnapshots,
    'forcemerge' : ForceMerge,
    'index_settings' : IndexSettings,
    'open' : Open,
    'reindex' : Reindex,
    'replicas' : Replicas,
    'restore' : Restore,
    'rollover' : Rollover,
    'snapshot' : Snapshot,
    'shrink' : Shrink,
}



"""Reindex action class"""
import logging
# pylint: disable=broad-except
from curator.exceptions import ActionError, ConfigurationError
from curator.helpers.getters import index_size, name_to_node_id, node_id_to_name, node_roles
from curator.helpers.testers import verify_index_list
from curator.helpers.utils import chunk_index_list, report_failure
from curator.helpers.waiters import health_check, wait_for_it


class Shrink:
    """Shrink Action Class"""
    def __init__(
            self, ilo, shrink_node='DETERMINISTIC', node_filters=None, number_of_shards=1,
            number_of_replicas=1, shrink_prefix='', shrink_suffix='-shrink', copy_aliases=False,
            delete_after=True, post_allocation=None, wait_for_active_shards=1,
            wait_for_rebalance=True, extra_settings=None, wait_for_completion=True, wait_interval=9,
            max_wait=-1
    ):
        """
        :param ilo: An IndexList Object
        :param shrink_node: The node name to use as the shrink target, or ``DETERMINISTIC``, which
            will use the values in ``node_filters`` to determine which node will be the shrink
            node.
        :param node_filters: If the value of ``shrink_node`` is ``DETERMINISTIC``, the values in
            ``node_filters`` will be used while determining which node to allocate the shards on
            before performing the shrink.
        :param number_of_shards: The number of shards the shrunk index should have
        :param number_of_replicas: The number of replicas for the shrunk index
        :param shrink_prefix: Prepend the shrunk index with this value
        :param shrink_suffix: Append the value to the shrunk index (Default: ``-shrink``)
        :param copy_aliases: Whether to copy each source index aliases to target index after
            shrinking. The aliases will be added to target index and deleted from source index at
            the same time. (Default: ``False``)
        :param delete_after: Whether to delete each index after shrinking. (Default: ``True``)
        :param post_allocation: If populated, the ``allocation_type``, ``key``, and ``value`` will
            be applied to the shrunk index to re-route it.
        :param extra_settings:  Permitted root keys are ``settings`` and ``aliases``.
        :param wait_for_active_shards: Wait for this many active shards before returning.
        :param wait_for_rebalance: Wait for rebalance. (Default: ``True``)
        :param wait_for_completion: Wait for completion before returning.
        :param wait_interval: Seconds to wait between completion checks.
        :param max_wait: Maximum number of seconds to ``wait_for_completion``

        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        :type shrink_node: str
        :type node_filters: dict
        :type number_of_shards: int
        :type number_of_replicas: int
        :type shrink_prefix: str
        :type shrink_suffix: str
        :type copy_aliases: bool
        :type delete_after: bool
        :type post_allocation: dict
        :type extra_settings: dict
        :type wait_for_active_shards: int
        :type wait_for_rebalance: bool
        :type wait_for_completion: bool
        :type wait_interval: int
        :type max_wait: int
        """
        if node_filters is None:
            node_filters = {}
        if post_allocation is None:
            post_allocation = {}
        if extra_settings is None:
            extra_settings = {}
        self.loggit = logging.getLogger('curator.actions.shrink')
        verify_index_list(ilo)
        if 'permit_masters' not in node_filters:
            node_filters['permit_masters'] = False
        #: The :py:class:`~.curator.indexlist.IndexList` object passed from param ``ilo``
        self.index_list = ilo
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`index_list`
        self.client = ilo.client
        #: Object attribute that gets the value of param ``shrink_node``.
        self.shrink_node = shrink_node
        #: Object attribute that gets the value of param ``node_filters``.
        self.node_filters = node_filters
        #: Object attribute that gets the value of param ``shrink_prefix``.
        self.shrink_prefix = shrink_prefix
        #: Object attribute that gets the value of param ``shrink_suffix``.
        self.shrink_suffix = shrink_suffix
        #: Object attribute that gets the value of param ``copy_aliases``.
        self.copy_aliases = copy_aliases
        #: Object attribute that gets the value of param ``delete_after``.
        self.delete_after = delete_after
        #: Object attribute that gets the value of param ``post_allocation``.
        self.post_allocation = post_allocation
        #: Object attribute that gets the value of param ``wait_for_rebalance``.
        self.wait_for_rebalance = wait_for_rebalance
        #: Object attribute that gets the value of param ``wait_for_completion``.
        self.wfc = wait_for_completion
        #: Object attribute that gets the value of param ``wait_interval``.
        self.wait_interval = wait_interval
        #: Object attribute that gets the value of param ``max_wait``.
        self.max_wait = max_wait
        #: Object attribute that gets the value of param ``number_of_shards``.
        self.number_of_shards = number_of_shards
        #: Object attribute that gets the value of param ``wait_for_active_shards``.
        self.wait_for_active_shards = wait_for_active_shards

        #: Object attribute that represents the target node for shrinking.
        self.shrink_node_name = None
        #: Object attribute that represents whether :py:attr:`shrink_node_name` is available
        self.shrink_node_avail = None
        #: Object attribute that represents the node_id of :py:attr:`shrink_node_name`
        self.shrink_node_id = None

        #: Object attribute that gets values from params ``number_of_shards`` and
        #: ``number_of_replicas``.
        self.settings = {
            'index.number_of_shards' : number_of_shards,
            'index.number_of_replicas' : number_of_replicas,
        }

        if extra_settings:
            self._merge_extra_settings(extra_settings)

        self._merge_extra_settings({
            'settings': {
                'index.routing.allocation.require._name': None,
                'index.blocks.write': None
            }})

    def _merge_extra_settings(self, extra_settings):
        self.loggit.debug('Adding extra_settings to shrink body: %s', extra_settings)
        # Pop these here, otherwise we could overwrite our default number of
        # shards and replicas
        if 'settings' in extra_settings:
            settings = extra_settings.pop('settings')
            try:
                self.settings.update(settings)
            except Exception as exc:
                raise ConfigurationError(
                    f"Unable to apply extra settings \"{{'settings':settings}}\" "
                    f"to shrink body. Exception: {exc}"
                ) from exc
        if extra_settings:
            try: # Apply any remaining keys, should there be any.
                self.settings.update(extra_settings)
            except Exception as exc:
                raise ConfigurationError(
                    f'Unable to apply extra settings "{extra_settings}" '
                    f'to shrink body. Exception: {exc}'
                ) from exc

    def _data_node(self, node_id):
        roles = node_roles(self.client, node_id)
        name = node_id_to_name(self.client, node_id)
        if not 'data' in roles:
            self.loggit.info('Skipping node "%s": non-data node', name)
            return False
        if 'master' in roles and not self.node_filters['permit_masters']:
            self.loggit.info('Skipping node "%s": master node', name)
            return False
        elif 'master' in roles and self.node_filters['permit_masters']:
            msg = (
                f'Not skipping node "{name}" which is a master node (not recommended), but '
                f'permit_masters is True'
            )
            self.loggit.warning(msg)
            return True
        else: # It does have `data` as a role.
            return True

    def _exclude_node(self, name):
        if 'exclude_nodes' in self.node_filters:
            if name in self.node_filters['exclude_nodes']:
                self.loggit.info('Excluding node "%s" due to node_filters', name)
                return True
        return False

    def _shrink_target(self, name):
        return f'{self.shrink_prefix}{name}{self.shrink_suffix}'

    def qualify_single_node(self):
        """Qualify a single node as a shrink target"""
        node_id = name_to_node_id(self.client, self.shrink_node)
        if node_id:
            self.shrink_node_id = node_id
            self.shrink_node_name = self.shrink_node
        else:
            raise ConfigurationError(
                f'Unable to find node named: "{self.shrink_node}"')
        if self._exclude_node(self.shrink_node):
            raise ConfigurationError(
                f'Node "{self.shrink_node}" listed for exclusion')
        if not self._data_node(node_id):
            raise ActionError(
                f'Node "{self.shrink_node}" is not usable as a shrink node')
        self.shrink_node_avail = (
            self.client.nodes.stats()['nodes'][node_id]['fs']['total']['available_in_bytes']
        )

    def most_available_node(self):
        """
        Determine which data node name has the most available free space, and meets the other node
        filters settings.
        """
        mvn_avail = 0
        # mvn_total = 0
        mvn_name = None
        mvn_id = None
        nodes = self.client.nodes.stats()['nodes']
        for node_id in nodes:
            name = nodes[node_id]['name']
            if self._exclude_node(name):
                self.loggit.debug('Node "%s" excluded by node filters', name)
                continue
            if not self._data_node(node_id):
                self.loggit.debug('Node "%s" is not a data node', name)
                continue
            value = nodes[node_id]['fs']['total']['available_in_bytes']
            if value > mvn_avail:
                mvn_name = name
                mvn_id = node_id
                mvn_avail = value
        self.shrink_node_name = mvn_name
        self.shrink_node_id = mvn_id
        self.shrink_node_avail = mvn_avail

    def route_index(self, idx, allocation_type, key, value):
        """Apply the indicated shard routing allocation"""
        bkey = f'index.routing.allocation.{allocation_type}.{key}'
        routing = {bkey : value}
        try:
            self.client.indices.put_settings(index=idx, settings=routing)
            if self.wait_for_rebalance:
                wait_for_it(
                    self.client, 'allocation', wait_interval=self.wait_interval,
                    max_wait=self.max_wait
                )
            else:
                wait_for_it(
                    self.client, 'relocate', index=idx, wait_interval=self.wait_interval,
                    max_wait=self.max_wait
                )
        except Exception as err:
            report_failure(err)

    def __log_action(self, error_msg, dry_run=False):
        if not dry_run:
            raise ActionError(error_msg)
        else:
            self.loggit.warning('DRY-RUN: %s', error_msg)

    def _block_writes(self, idx):
        block = {'index.blocks.write': True}
        self.client.indices.put_settings(index=idx, settings=block)

    def _unblock_writes(self, idx):
        unblock = {'index.blocks.write': False}
        self.client.indices.put_settings(index=idx, settings=unblock)

    def _check_space(self, idx, dry_run=False):
        # Disk watermark calculation is already baked into `available_in_bytes`
        size = index_size(self.client, idx, value='primaries')
        padded = (size * 2) + (32 * 1024)
        if padded < self.shrink_node_avail:
            msg = (
                f'Sufficient space available for 2x the size of index "{idx}". '
                f'Required: {padded}, available: {self.shrink_node_avail}'
            )
            self.loggit.debug(msg)
        else:
            error_msg = (
                f'Insufficient space available for 2x the size of index "{idx}", shrinking will '
                f'exceed space available. Required: {padded}, available: {self.shrink_node_avail}'
            )
            self.__log_action(error_msg, dry_run)

    def _check_node(self):
        if self.shrink_node != 'DETERMINISTIC':
            if not self.shrink_node_name:
                self.qualify_single_node()
        else:
            self.most_available_node()
        # At this point, we should have the three shrink-node identifying
        # instance variables:
        # - self.shrink_node_name
        # - self.shrink_node_id
        # - self.shrink_node_avail
        # # - self.shrink_node_total - only if needed in the future

    def _check_target_exists(self, idx, dry_run=False):
        target = self._shrink_target(idx)
        if self.client.indices.exists(index=target):
            error_msg = f'Target index "{target}" already exists'
            self.__log_action(error_msg, dry_run)

    def _check_doc_count(self, idx, dry_run=False):
        max_docs = 2147483519
        doc_count = self.client.indices.stats(index=idx)['indices'][idx]['primaries']['docs']['count']
        if doc_count > (max_docs * self.number_of_shards):
            error_msg = (
                f'Too many documents ({doc_count}) to fit in {self.number_of_shards} shard(s). '
                f'Maximum number of docs per shard is {max_docs}'
            )
            self.__log_action(error_msg, dry_run)

    def _check_shard_count(self, idx, src_shards, dry_run=False):
        if self.number_of_shards >= src_shards:
            error_msg = (
                f'Target number of shards ({self.number_of_shards}) must be less than current '
                f'number of shards ({src_shards}) in index "{idx}"'
            )
            self.__log_action(error_msg, dry_run)

    def _check_shard_factor(self, idx, src_shards, dry_run=False):
        # Find the list of factors of src_shards
        factors = [x for x in range(1, src_shards+1) if src_shards % x == 0]
        # Pop the last one, because it will be the value of src_shards
        factors.pop()
        if not self.number_of_shards in factors:
            error_msg = (
                f'"{self.number_of_shards}" is not a valid factor of {src_shards} shards of '
                f'index {idx}. Valid values are {factors}'
            )
            self.__log_action(error_msg, dry_run)

    def _check_all_shards(self, idx):
        shards = self.client.cluster.state(index=idx)['routing_table']['indices'][idx]['shards']
        found = []
        for shardnum in shards:
            for shard_idx in range(0, len(shards[shardnum])):
                if shards[shardnum][shard_idx]['node'] == self.shrink_node_id:
                    found.append(
                        {'shard': shardnum, 'primary': shards[shardnum][shard_idx]['primary']})
        if len(shards) != len(found):
            self.loggit.debug(
                'Found these shards on node "%s": %s', self.shrink_node_name, found)
            raise ActionError(
                f'Unable to shrink index "{idx}" as not all shards were found on the designated '
                f'shrink node ({self.shrink_node_name}): {found}'
            )

    def pre_shrink_check(self, idx, dry_run=False):
        """Do a shrink preflight check"""
        self.loggit.debug('BEGIN PRE_SHRINK_CHECK')
        self.loggit.debug('Check that target exists')
        self._check_target_exists(idx, dry_run)
        self.loggit.debug('Check doc count constraints')
        self._check_doc_count(idx, dry_run)
        self.loggit.debug('Check shard count')
        src_shards = int(self.client.indices.get(index=idx)[idx]['settings']['index']['number_of_shards'])
        self._check_shard_count(idx, src_shards, dry_run)
        self.loggit.debug('Check shard factor')
        self._check_shard_factor(idx, src_shards, dry_run)
        self.loggit.debug('Check node availability')
        self._check_node()
        self.loggit.debug('Check available disk space')
        self._check_space(idx, dry_run)
        self.loggit.debug('FINISH PRE_SHRINK_CHECK')

    def do_copy_aliases(self, source_idx, target_idx):
        """Copy the aliases to the shrunk index"""
        alias_actions = []
        aliases = self.client.indices.get_alias(index=source_idx)
        for alias in aliases[source_idx]['aliases']:
            self.loggit.debug('alias: %s', alias)
            alias_actions.append({'remove': {'index': source_idx, 'alias': alias}})
            alias_actions.append({'add': {'index': target_idx, 'alias': alias}})
        if alias_actions:
            self.loggit.info('Copy alias actions: %s', alias_actions)
            self.client.indices.update_aliases(actions=alias_actions)

    def do_dry_run(self):
        """Show what a regular run would do, but don't actually do it."""
        self.index_list.filter_closed()
        self.index_list.filter_by_shards(number_of_shards=self.number_of_shards)
        self.index_list.empty_list_check()
        try:
            index_lists = chunk_index_list(self.index_list.indices)
            for lst in index_lists:
                for idx in lst: # Shrink can only be done one at a time...
                    target = self._shrink_target(idx)
                    self.pre_shrink_check(idx, dry_run=True)
                    self.loggit.info(
                        'DRY-RUN: Moving shards to shrink node: "%s"', self.shrink_node_name)
                    msg = (
                        f'DRY-RUN: Shrinking index "{idx}" to "{target}" with settings: '
                        f'{self.settings}, wait_for_active_shards={self.wait_for_active_shards}'
                    )
                    self.loggit.info(msg)
                    if self.post_allocation:
                        submsg = (
                            f"index.routing.allocation.{self.post_allocation['allocation_type']}."
                            f"{self.post_allocation['key']}:{self.post_allocation['value']}"
                        )
                        msg = (
                            f'DRY-RUN: Applying post-shrink allocation rule "{submsg}" to index '
                            f'"{target}"'
                        )
                        self.loggit.info(msg)
                    if self.copy_aliases:
                        msg = (
                            f'DRY-RUN: Copy source index aliases '
                            f'"{self.client.indices.get_alias(index=idx)}"'
                        )
                        self.loggit.info(msg)
                    if self.delete_after:
                        self.loggit.info('DRY-RUN: Deleting source index "%s"', idx)
        except Exception as err:
            report_failure(err)

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.IndicesClient.shrink` the indices in :py:attr:`index_list`
        """
        self.index_list.filter_closed()
        self.index_list.filter_by_shards(number_of_shards=self.number_of_shards)
        self.index_list.empty_list_check()
        msg = (
            f'Shrinking {len(self.index_list.indices)} selected indices: {self.index_list.indices}'
        )
        self.loggit.info(msg)
        try:
            index_lists = chunk_index_list(self.index_list.indices)
            for lst in index_lists:
                for idx in lst: # Shrink can only be done one at a time...
                    target = self._shrink_target(idx)
                    self.loggit.info('Source index: %s -- Target index: %s', idx, target)
                    # Pre-check ensures disk space available for each pass of the loop
                    self.pre_shrink_check(idx)
                    # Route the index to the shrink node
                    self.loggit.info('Moving shards to shrink node: "%s"', self.shrink_node_name)
                    self.route_index(idx, 'require', '_name', self.shrink_node_name)
                    # Ensure a copy of each shard is present
                    self._check_all_shards(idx)
                    # Block writes on index
                    self._block_writes(idx)
                    # Do final health check
                    if not health_check(self.client, status='green'):
                        raise ActionError(
                            'Unable to proceed with shrink action. Cluster health is not "green"')
                    # Do the shrink
                    msg = (
                        f'Shrinking index "{idx}" to "{target}" with settings: {self.settings}, '
                        f'wait_for_active_shards={self.wait_for_active_shards}'
                    )
                    self.loggit.info(msg)
                    try:
                        self.client.indices.shrink(
                            index=idx, target=target, settings=self.settings,
                            wait_for_active_shards=self.wait_for_active_shards
                        )
                        # Wait for it to complete
                        if self.wfc:
                            self.loggit.debug(
                                'Wait for shards to complete allocation for index: %s', target)
                            if self.wait_for_rebalance:
                                wait_for_it(
                                    self.client, 'shrink', wait_interval=self.wait_interval,
                                    max_wait=self.max_wait
                                )
                            else:
                                wait_for_it(
                                    self.client, 'relocate', index=target,
                                    wait_interval=self.wait_interval, max_wait=self.max_wait
                                )
                    except Exception as exc:
                        if self.client.indices.exists(index=target):
                            msg = (
                                f'Deleting target index "{target}" due to failure to complete '
                                f'shrink'
                            )
                            self.loggit.error(msg)
                            self.client.indices.delete(index=target)
                        raise ActionError(
                            f'Unable to shrink index "{idx}" -- Error: {exc}') from exc
                    self.loggit.info('Index "%s" successfully shrunk to "%s"', idx, target)
                    # Do post-shrink steps
                    # Unblock writes on index (just in case)
                    self._unblock_writes(idx)
                    ## Post-allocation, if enabled
                    if self.post_allocation:
                        submsg = (
                            f"index.routing.allocation.{self.post_allocation['allocation_type']}."
                            f"{self.post_allocation['key']}:{self.post_allocation['value']}"
                        )
                        msg = (
                            f'Applying post-shrink allocation rule "{submsg}" to index "{target}"'
                        )
                        self.loggit.info(msg)
                        self.route_index(
                            target, self.post_allocation['allocation_type'],
                            self.post_allocation['key'], self.post_allocation['value']
                        )
                    ## Copy aliases, if flagged
                    if self.copy_aliases:
                        self.loggit.info('Copy source index aliases "%s"', idx)
                        self.do_copy_aliases(idx, target)
                    ## Delete, if flagged
                    if self.delete_after:
                        self.loggit.info('Deleting source index "%s"', idx)
                        self.client.indices.delete(index=idx)
                    else: # Let's unset the routing we applied here.
                        self.loggit.info('Unassigning routing for source index: "%s"', idx)
                        self.route_index(idx, 'require', '_name', '')

        except Exception as err:
            # Just in case it fails after attempting to meet this condition
            self._unblock_writes(idx)
            report_failure(err)



"""Delete Index and Delete Snapshot Singletons"""
import click
from curator.cli_singletons.object_class import CLIAction
from curator.cli_singletons.utils import get_width, validate_filter_json

#### Indices ####
@click.command(context_settings=get_width())
@click.option(
    '--ignore_empty_list',
    is_flag=True,
    help='Do not raise exception if there are no actionable indices'
)
@click.option(
    '--allow_ilm_indices/--no-allow_ilm_indices',
    help='Allow Curator to operate on Index Lifecycle Management monitored indices.',
    default=False,
    show_default=True
)
@click.option(
    '--filter_list',
    callback=validate_filter_json,
    help='JSON array of filters selecting indices to act on.',
    required=True
)
@click.pass_context
def delete_indices(ctx, ignore_empty_list, allow_ilm_indices, filter_list):
    """
    Delete Indices
    """
    # ctx.info_name is the name of the function or name specified in @click.command decorator
    action = CLIAction(
        'delete_indices',
        ctx.obj['config'],
        {'allow_ilm_indices':allow_ilm_indices},
        filter_list,
        ignore_empty_list
    )
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])

#### Snapshots ####
@click.command(context_settings=get_width())
@click.option('--repository', type=str, required=True, help='Snapshot repository name')
@click.option('--retry_count', type=int, help='Number of times to retry (max 3)')
@click.option('--retry_interval', type=int, help='Time in seconds between retries')
@click.option(
    '--ignore_empty_list',
    is_flag=True,
    help='Do not raise exception if there are no actionable snapshots'
)
@click.option(
    '--allow_ilm_indices/--no-allow_ilm_indices',
    help='Allow Curator to operate on Index Lifecycle Management monitored indices.',
    default=False,
    show_default=True
)
@click.option(
    '--filter_list',
    callback=validate_filter_json,
    help='JSON array of filters selecting snapshots to act on.',
    required=True
)
@click.pass_context
def delete_snapshots(
        ctx, repository, retry_count, retry_interval,
        ignore_empty_list, allow_ilm_indices, filter_list
    ):
    """
    Delete Snapshots
    """
    manual_options = {
        'retry_count': retry_count,
        'retry_interval': retry_interval,
        'allow_ilm_indices': allow_ilm_indices,
    }
    # ctx.info_name is the name of the function or name specified in @click.command decorator
    action = CLIAction(
        'delete_snapshots',
        ctx.obj['config'],
        manual_options,
        filter_list,
        ignore_empty_list,
        repository=repository
    )
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])



"""The function that waits

...and its helpers
"""
import logging
from time import localtime, sleep, strftime
from datetime import datetime
from curator.exceptions import (
    ActionTimeout, ConfigurationError, CuratorException, FailedReindex, MissingArgument)
from curator.helpers.utils import chunk_index_list

def health_check(client, **kwargs):
    """
    This function calls `client.cluster.` :py:meth:`~.elasticsearch.client.ClusterClient.health`
    and, based on the params provided, will return ``True`` or ``False`` depending on whether that
    particular keyword appears in the output, and has the expected value.

    If multiple keys are provided, all must match for a ``True`` response.

    :param client: A client connection object

    :type client: :py:class:`~.elasticsearch.Elasticsearch`

    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    logger.debug('KWARGS= "%s"', kwargs)
    klist = list(kwargs.keys())
    if not klist:
        raise MissingArgument('Must provide at least one keyword argument')
    hc_data = client.cluster.health()
    response = True

    for k in klist:
        # First, verify that all kwargs are in the list
        if not k in list(hc_data.keys()):
            raise ConfigurationError('Key "{0}" not in cluster health output')
        if not hc_data[k] == kwargs[k]:
            msg = f'NO MATCH: Value for key "{kwargs[k]}", health check data: {hc_data[k]}'
            logger.debug(msg)
            response = False
        else:
            msg = f'MATCH: Value for key "{kwargs[k]}", health check data: {hc_data[k]}'
            logger.debug(msg)
    if response:
        logger.info('Health Check for all provided keys passed.')
    return response

def relocate_check(client, index):
    """
    This function calls `client.cluster.` :py:meth:`~.elasticsearch.client.ClusterClient.state`
    with a given index to check if all of the shards for that index are in the ``STARTED`` state.
    It will return ``True`` if all primary and replica shards are in the ``STARTED`` state, and it
    will return ``False`` if any shard is in a different state.

    :param client: A client connection object
    :param index: The index name

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type index: str

    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    shard_state_data = (
        client.cluster.state(index=index)['routing_table']['indices'][index]['shards']
    )
    finished_state = (
        all(
            all(
                shard['state'] == "STARTED" for shard in shards
            )
            for shards in shard_state_data.values()
        )
    )
    if finished_state:
        logger.info('Relocate Check for index: "%s" has passed.', index)
    return finished_state

def restore_check(client, index_list):
    """
    This function calls `client.indices.` :py:meth:`~.elasticsearch.client.IndicesClient.recovery`
    with the list of indices to check for complete recovery.  It will return ``True`` if recovery
    of those indices is complete, and ``False`` otherwise.  It is designed to fail fast: if a
    single shard is encountered that is still recovering (not in ``DONE`` stage), it will
    immediately return ``False``, rather than complete iterating over the rest of the response.

    :param client: A client connection object
    :param index_list: The list of indices to verify having been restored.

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type index_list: list

    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    response = {}
    for chunk in chunk_index_list(index_list):
        try:
            chunk_response = client.indices.recovery(index=chunk, human=True)
        except Exception as err:
            msg = f'Unable to obtain recovery information for specified indices. Error: {err}'
            raise CuratorException(msg) from err
        if chunk_response == {}:
            logger.info('_recovery returned an empty response. Trying again.')
            return False
        response.update(chunk_response)
    logger.info('Provided indices: %s', index_list)
    logger.info('Found indices: %s', list(response.keys()))
    # pylint: disable=consider-using-dict-items
    for index in response:
        for shard in range(0, len(response[index]['shards'])):
            stage = response[index]['shards'][shard]['stage']
            if stage != 'DONE':
                logger.info('Index "%s" is still in stage "%s"', index, stage)
                return False

    # If we've gotten here, all of the indices have recovered
    return True

def snapshot_check(client, snapshot=None, repository=None):
    """
    This function calls `client.snapshot.` :py:meth:`~.elasticsearch.client.SnapshotClient.get` and
    tests to see whether the snapshot is complete, and if so, with what status.  It will log errors
    according to the result. If the snapshot is still ``IN_PROGRESS``, it will return ``False``.
    ``SUCCESS`` will be an ``INFO`` level message, ``PARTIAL`` nets a ``WARNING`` message,
    ``FAILED`` is an ``ERROR``, message, and all others will be a ``WARNING`` level message.

    :param client: A client connection object
    :param snapshot: The snapshot name
    :param repository: The repository name

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type snapshot: str
    :type repository: str

    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    logger.debug('SNAPSHOT: %s', snapshot)
    logger.debug('REPOSITORY: %s', repository)
    try:
        result = client.snapshot.get(repository=repository, snapshot=snapshot)
        logger.debug('RESULT: %s', result)
    except Exception as err:
        raise CuratorException(
            f'Unable to obtain information for snapshot "{snapshot}" in repository '
            f'"{repository}". Error: {err}'
        ) from err
    state = result['snapshots'][0]['state']
    logger.debug('Snapshot state = %s', state)
    retval = True
    if state == 'IN_PROGRESS':
        logger.info('Snapshot %s still in progress.', snapshot)
        retval = False
    elif state == 'SUCCESS':
        logger.info('Snapshot %s successfully completed.', snapshot)
    elif state == 'PARTIAL':
        logger.warning('Snapshot %s completed with state PARTIAL.', snapshot)
    elif state == 'FAILED':
        logger.error('Snapshot %s completed with state FAILED.', snapshot)
    else:
        logger.warning('Snapshot %s completed with state: %s', snapshot, state)
    return retval

def task_check(client, task_id=None):
    """
    This function calls `client.tasks.` :py:meth:`~.elasticsearch.client.TasksClient.get` with the
    provided ``task_id``.  If the task data contains ``'completed': True``, then it will return
    ``True``. If the task is not completed, it will log some information about the task and return
    ``False``

    :param client: A client connection object
    :param task_id: The task id

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type task_id: str

    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    try:
        task_data = client.tasks.get(task_id=task_id)
    except Exception as err:
        msg = f'Unable to obtain task information for task_id "{task_id}". Exception {err}'
        raise CuratorException(msg) from err
    task = task_data['task']
    completed = task_data['completed']
    if task['action'] == 'indices:data/write/reindex':
        logger.debug('It\'s a REINDEX TASK')
        logger.debug('TASK_DATA: %s', task_data)
        logger.debug('TASK_DATA keys: %s', list(task_data.keys()))
        if 'response' in task_data:
            response = task_data['response']
            if response['failures']:
                msg = f'Failures found in reindex response: {response["failures"]}'
                raise FailedReindex(msg)
    running_time = 0.000000001 * task['running_time_in_nanos']
    logger.debug('Running time: %s seconds', running_time)
    descr = task['description']

    if completed:
        completion_time = (running_time * 1000) + task['start_time_in_millis']
        time_string = strftime('%Y-%m-%dT%H:%M:%SZ', localtime(completion_time/1000))
        logger.info('Task "%s" completed at %s.', descr, time_string)
        retval = True
    else:
        # Log the task status here.
        logger.debug('Full Task Data: %s', task_data)
        msg = (
            f'Task "{descr}" with task_id "{task_id}" has been running for {running_time} seconds'
        )
        logger.info(msg)
        retval = False
    return retval

# pylint: disable=too-many-locals, too-many-arguments
def wait_for_it(
        client, action, task_id=None, snapshot=None, repository=None, index=None, index_list=None,
        wait_interval=9, max_wait=-1
    ):
    """
    This function becomes one place to do all ``wait_for_completion`` type behaviors

    :param client: A client connection object
    :param action: The action name that will identify how to wait
    :param task_id: If the action provided a task_id, this is where it must be declared.
    :param snapshot: The name of the snapshot.
    :param repository: The Elasticsearch snapshot repository to use
    :param wait_interval: Seconds to wait between completion checks.
    :param max_wait: Maximum number of seconds to ``wait_for_completion``

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type action: str
    :type task_id: str
    :type snapshot: str
    :type repository: str
    :type wait_interval: int
    :type max_wait: int
    :rtype: None
    """
    logger = logging.getLogger(__name__)
    action_map = {
        'allocation':{'function': health_check, 'args': {'relocating_shards':0}},
        'replicas':{'function': health_check, 'args': {'status':'green'}},
        'cluster_routing':{'function': health_check, 'args': {'relocating_shards':0}},
        'snapshot':{
            'function':snapshot_check, 'args':{'snapshot':snapshot, 'repository':repository}},
        'restore':{'function':restore_check, 'args':{'index_list':index_list}},
        'reindex':{'function':task_check, 'args':{'task_id':task_id}},
        'shrink':{'function': health_check, 'args': {'status':'green'}},
        'relocate':{'function': relocate_check, 'args': {'index':index}},
    }
    wait_actions = list(action_map.keys())

    if action not in wait_actions:
        raise ConfigurationError(f'"action" must be one of {wait_actions}')
    if action == 'reindex' and task_id is None:
        raise MissingArgument(f'A task_id must accompany "action" {action}')
    if action == 'snapshot' and ((snapshot is None) or (repository is None)):
        raise MissingArgument(
            f'A snapshot and repository must accompany "action" {action}. snapshot: '
            f'{snapshot}, repository: {repository}'
        )
    if action == 'restore' and index_list is None:
        raise MissingArgument(f'An index_list must accompany "action" {action}')
    if action == 'reindex':
        try:
            _ = client.tasks.get(task_id=task_id)
        except Exception as err:
            # This exception should only exist in API usage. It should never
            # occur in regular Curator usage.
            raise CuratorException(f'Unable to find task_id {task_id}. Exception: {err}') from err

    # Now with this mapped, we can perform the wait as indicated.
    start_time = datetime.now()
    result = False
    while True:
        elapsed = int((datetime.now() - start_time).total_seconds())
        logger.debug('Elapsed time: %s seconds', elapsed)
        response = action_map[action]['function'](client, **action_map[action]['args'])
        logger.debug('Response: %s', response)
        # Success
        if response:
            logger.debug(
                'Action "%s" finished executing (may or may not have been successful)', action)
            result = True
            break
        # Not success, and reached maximum wait (if defined)
        if (max_wait != -1) and (elapsed >= max_wait):
            msg = f'Unable to complete action "{action}" within max_wait ({max_wait}) seconds.'
            logger.error(msg)
            break
        # Not success, so we wait.
        msg = (
            f'Action "{action}" not yet complete, {elapsed} total seconds elapsed. '
            f'Waiting {wait_interval} seconds before checking again.'
        )
        logger.debug(msg)
        sleep(wait_interval)

    logger.debug('Result: %s', result)
    if not result:
        raise ActionTimeout(
            f'Action "{action}" failed to complete in the max_wait period of {max_wait} seconds'
        )



"""Use __init__ to make these not need to be nested under lowercase.Capital"""
from curator.cli_singletons.alias import alias
from curator.cli_singletons.allocation import allocation
from curator.cli_singletons.close import close
from curator.cli_singletons.delete import delete_indices, delete_snapshots
from curator.cli_singletons.forcemerge import forcemerge
from curator.cli_singletons.open_indices import open_indices
from curator.cli_singletons.replicas import replicas
from curator.cli_singletons.restore import restore
from curator.cli_singletons.rollover import rollover
from curator.cli_singletons.shrink import shrink
from curator.cli_singletons.snapshot import snapshot



"""Close index action class"""
import logging
from curator.helpers.testers import verify_index_list
from curator.helpers.utils import chunk_index_list, report_failure, show_dry_run, to_csv

class Close:
    """Close Action Class"""
    def __init__(self, ilo, delete_aliases=False, skip_flush=False):
        """
        :param ilo: An IndexList Object
        :param delete_aliases: Delete any associated aliases before closing indices.
        :param skip_flush: Do not flush indices before closing.

        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        :type delete_aliases: bool
        :type skip_flush: bool
        """
        verify_index_list(ilo)
        #: The :py:class:`~.curator.indexlist.IndexList` object passed from param ``ilo``
        self.index_list = ilo
        #: The value passed as ``delete_aliases``
        self.delete_aliases = delete_aliases
        #: The value passed as ``skip_flush``
        self.skip_flush = skip_flush
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`index_list`
        self.client = ilo.client
        self.loggit = logging.getLogger('curator.actions.close')

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        show_dry_run(self.index_list, 'close', **{'delete_aliases':self.delete_aliases})

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.IndicesClient.close` open indices in :py:attr:`index_list`
        """
        self.index_list.filter_closed()
        self.index_list.empty_list_check()
        self.loggit.info(
            'Closing %s selected indices: %s',
                len(self.index_list.indices), self.index_list.indices
        )
        try:
            index_lists = chunk_index_list(self.index_list.indices)
            for lst in index_lists:
                lst_as_csv = to_csv(lst)
                self.loggit.debug('CSV list of indices to close:  %s', lst_as_csv)
                if self.delete_aliases:
                    self.loggit.info('Deleting aliases from indices before closing.')
                    self.loggit.debug('Deleting aliases from:  %s', lst)
                    try:
                        self.client.indices.delete_alias(index=lst_as_csv, name='_all')
                        self.loggit.debug('Deleted aliases from: %s', lst)
                    # pylint: disable=broad-except
                    except Exception as err:
                        self.loggit.warning(
                            'Some indices may not have had aliases.  Exception: %s', err)
                if not self.skip_flush:
                    self.client.indices.flush(index=lst_as_csv, ignore_unavailable=True, force=True)
                self.client.indices.close(index=lst_as_csv, ignore_unavailable=True)
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)
